#ifndef _TYPES_H_
#define _TYPES_H_

#include <Eigen/Core>
#include <vector>
#include <iostream>

#define LOG(X) std::cout << X << std::endl;

namespace MPC {

        typedef Eigen::MatrixXf MatrixT;
        typedef Eigen::VectorXf VectorT;
        
        struct DDPParams {
                float zMin;
                float lamInit;
                float lambdaFactor;
                float lambdaMax;
                float lambdaMin;
                float tolFun;
                float tolGrad;
                float dlamInit;
                
                DDPParams() {
                        zMin =  0.2;
                        lamInit = 1;
                        lambdaFactor = 1.2;
                        lambdaMax = 1e10;
                        lambdaMin = 1e-6;
                        tolFun = 1e-9;
                        tolGrad = 1e-6;
                        dlamInit = 1;
                }
                
        };
        
        struct Trajectory {
                EIGEN_MAKE_ALIGNED_OPERATOR_NEW;

                MatrixT x;
                MatrixT u;
                VectorT cost;
                bool started;

                Trajectory() {
                        started = false;
                }
        };

        struct Deriv {
                EIGEN_MAKE_ALIGNED_OPERATOR_NEW;

                MatrixT fx;
                MatrixT fu;
                std::vector<MatrixT> fxx;
                std::vector<MatrixT> fxu;
                std::vector<MatrixT> fuu;
                VectorT cx;
                VectorT cu;
                MatrixT cxx;
                MatrixT cxu;
                MatrixT cuu;
        };
        
        struct LocalValueFunction {
                EIGEN_MAKE_ALIGNED_OPERATOR_NEW;

                MatrixT V; // TODO: we do not actually compute this right now do we ?
                MatrixT Vx;
                std::vector<MatrixT> Vxx;
                VectorT dV;

                MatrixT l;
                std::vector<MatrixT> L;
        };


        typedef std::vector<Deriv> TrajectoryDeriv;

        typedef std::function<float(const VectorT &, const VectorT &)> CostFunction;
        typedef std::function<void(const VectorT &, const VectorT &, Deriv &)> CostFunctionD;
        typedef std::function<void(const VectorT &, const VectorT &, VectorT &)> Dynamics;
        typedef std::function<void(const VectorT &, const VectorT &, Deriv &)> DynamicsD;

}

#endif
